/*
 * drivers/usb/host/ehci-suh.c
 *
 * USB controller driver
 *
 * Copyright 2012 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Suite 500, Boston, MA 02110-1335 USA.
 *
 */

/*--------------------------------------------------------------------------*
 * Include File
 *---------------------------------------------------------------------------*/
#include <linux/platform_device.h>
#include "ehci-suh.h"
#ifdef CONFIG_POSTEDWRITE_WORKAROUND
#include "pos-suh_def.h"
#include <mach/postedwrite_sync.h>
#endif

/*--------------------------------------------------------------------------*
 * Function Prototype Declare
 *---------------------------------------------------------------------------*/
static void ehci_suh_start(struct platform_device *device);
static void ehci_suh_stop(struct platform_device *device);
static void ehci_suh_set_register(struct platform_device *device, u32 opbase);
static int ehci_suh_setup(struct usb_hcd *hcd);
static int ehci_suh_drv_probe(struct platform_device *device);
static int ehci_suh_drv_remove(struct platform_device *device);
#ifdef CONFIG_PM
static int ehci_suh_drv_suspend(struct platform_device *device, pm_message_t message);
static int ehci_suh_drv_resume(struct platform_device *device);
#endif

/*--------------------------------------------------------------------------*
 * Variable Declare
 *---------------------------------------------------------------------------*/
/*
 * hc_driver
 */
static const struct hc_driver ehci_suh_hc_driver = {
	.description		= hcd_name,
	.hcd_priv_size		= sizeof(struct ehci_hcd),
	.irq			= ehci_irq,
	.flags			= HCD_MEMORY | HCD_USB2,
	.reset			= ehci_suh_setup,
	.start			= ehci_run,
	.stop			= ehci_stop,
	.shutdown		= ehci_shutdown,
	.urb_enqueue		= ehci_urb_enqueue,
	.urb_dequeue		= ehci_urb_dequeue,
	.endpoint_disable	= ehci_endpoint_disable,
	.endpoint_reset		= ehci_endpoint_reset,
	.get_frame_number	= ehci_get_frame,
	.hub_status_data	= ehci_hub_status_data,
	.hub_control		= ehci_hub_control,
#ifdef  CONFIG_PM
	.bus_suspend		= ehci_bus_suspend,
	.bus_resume		= ehci_bus_resume,
#endif
	.relinquish_port	= ehci_relinquish_port,
	.port_handed_over	= ehci_port_handed_over,
	.clear_tt_buffer_complete = ehci_clear_tt_buffer_complete,
};

/*
 * platform_driver
 */
static struct platform_driver ehci_hcd_suh_driver = {
	.probe		= ehci_suh_drv_probe,
	.remove		= ehci_suh_drv_remove,
	.shutdown	= usb_hcd_platform_shutdown,
#ifdef  CONFIG_PM
	.suspend	= ehci_suh_drv_suspend,
	.resume		= ehci_suh_drv_resume,
#endif
	.driver		= {
		.name	= "ehci-suh",
		.owner	= THIS_MODULE,
	}
};

/*--------------------------------------------------------------------------*
 * For Debug
 *---------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------*
 * System Depend
 *---------------------------------------------------------------------------*/
/*
 * ehci_suh_start
 */
static void ehci_suh_start(struct platform_device *device)
{
	dev_dbg(&device->dev,"ehci_suh_start\n");
	return;
}

/*
 * ehci_suh_stop
 */
static void ehci_suh_stop(struct platform_device *device)
{
	dev_dbg(&device->dev,"ehci_suh_stop\n");
	return;
}

/*
 * ehci_suh_set_register
 */
static void ehci_suh_set_register(struct platform_device *device, u32 opbase)
{
	iowrite32(0x01000100, (opbase + SUH_OFFSET_INSNREG01) );

	dev_dbg(&device->dev,"SUH_INSNREG00=%X\n",
			ioread32((void*)(opbase + SUH_OFFSET_INSNREG00)) );
	dev_dbg(&device->dev,"SUH_INSNREG01=%X\n",
			ioread32((void*)(opbase + SUH_OFFSET_INSNREG01)) );
	dev_dbg(&device->dev,"SUH_INSNREG02=%X\n",
			ioread32((void*)(opbase + SUH_OFFSET_INSNREG02)) );
	dev_dbg(&device->dev,"SUH_INSNREG03=%X\n",
			ioread32((void*)(opbase + SUH_OFFSET_INSNREG03)) );
	dev_dbg(&device->dev,"SUH_INSNREG04=%X\n",
			ioread32((void*)(opbase + SUH_OFFSET_INSNREG04)) );
	dev_dbg(&device->dev,"SUH_INSNREG05=%X\n",
			ioread32((void*)(opbase + SUH_OFFSET_INSNREG05)) );
	dev_dbg(&device->dev,"SUH_INSNREG06=%X\n",
			ioread32((void*)(opbase + SUH_OFFSET_INSNREG06)) );
	dev_dbg(&device->dev,"SUH_INSNREG07=%X\n",
			ioread32((void*)(opbase + SUH_OFFSET_INSNREG07)) );
	dev_dbg(&device->dev,"SUH_INSNREG08=%X\n",
			ioread32((void*)(opbase + SUH_OFFSET_INSNREG08)) );
	return;
}

/*--------------------------------------------------------------------------*
 * hc_driver suh depend
 *---------------------------------------------------------------------------*/
static int ehci_suh_setup(struct usb_hcd *hcd)
{
	struct ehci_hcd *ehci;
	int ret;

	ehci = hcd_to_ehci(hcd);
	ehci->caps = hcd->regs;
	ehci->regs = hcd->regs +
			HC_LENGTH(ehci_readl(ehci, &ehci->caps->hc_capbase));
	ehci->hcs_params = ehci_readl(ehci, &ehci->caps->hcs_params);

	ret = ehci_halt(ehci);
	if ( ret != 0 ) {
		return ret;
	}

	ret = ehci_init(hcd);
#ifdef USB_SUH_PW
	postedwrite_sync( POSTEDWRITE_DRAM );
#endif
	if ( ret != 0 ) {
		return ret;
	}

	ret = ehci_reset(ehci);
	if ( ret != 0 ) {
		return ret;
	}

	ehci_port_power(ehci, 1);

	return 0;
}

/*--------------------------------------------------------------------------*
 * platform_driver suh depend
 *---------------------------------------------------------------------------*/
/*
 * .probe
 */
static int ehci_suh_drv_probe(struct platform_device *device)
{
	struct usb_hcd *hcd;
	struct ehci_hcd *ehci;
	struct resource *res;
	struct resource *mem;
	int irq;
	void __iomem *iomem;
	int ret;

	if ( usb_disabled() ) {
		return -ENODEV;
	}

	res = platform_get_resource(device, IORESOURCE_IRQ, 0);
	if ( res == 0 ) {
		dev_dbg(&device->dev,"platform_get_resource IRQ fail\n");
		return -ENODEV;
	}
	irq = res->start;

	res = platform_get_resource(device, IORESOURCE_MEM, 0);
	if ( res == 0 ) {
		dev_dbg(&device->dev,"platform_get_resource MEM fail\n");
		return -ENODEV;
	}

	mem = request_mem_region(res->start, res->end - res->start + 1,
					ehci_suh_hc_driver.description);
	if ( mem == 0 ) {
		dev_dbg(&device->dev,"request_mem_region fail\n");
		return -EBUSY;
	}

	iomem = ioremap_nocache(res->start, res->end - res->start + 1);
	if ( iomem == 0 ) {
		release_mem_region(res->start, res->end - res->start + 1);
		dev_dbg(&device->dev,"ioremap fail\n");
		return -EFAULT;
	}

	hcd = usb_create_hcd(&ehci_suh_hc_driver,
				&device->dev,
				dev_name(&device->dev));
	if ( hcd == 0 ) {
		iounmap(iomem);
		release_mem_region(res->start, res->end - res->start + 1);
		dev_dbg(&device->dev,"usb_create_hcd fail\n");
		return -ENOMEM;
	}
	hcd->rsrc_start = res->start;
	hcd->rsrc_len = res->end - res->start + 1;
	hcd->regs = iomem;
	ehci = hcd_to_ehci(hcd);
	ehci->caps = hcd->regs;
	ehci->regs = hcd->regs + HC_LENGTH(ioread32(&ehci->caps->hc_capbase));
	ehci->hcs_params = ioread32(&ehci->caps->hcs_params);

	dev_dbg(&device->dev,"IRQ=%X MEM start=%X end=%X ioremap=%X\n",
				(int)irq,
				(int)res->start,(int)res->end,
				(int)iomem);

	/* System Depend Process Start */
	ehci_suh_start(device);
	ehci_suh_set_register(device,(u32)ehci->regs);
	/* System Depend Process End */

	ret = usb_add_hcd(hcd, irq, IRQF_DISABLED | IRQF_SHARED);
	if ( ret != 0 ) {
		usb_put_hcd(hcd);
		iounmap(iomem);
		release_mem_region(res->start, res->end - res->start + 1);
		dev_dbg(&device->dev,"usb_add_hcd fail\n");
		return ret;
	}

	platform_set_drvdata(device, hcd);

	dev_dbg(&device->dev,"ehci_suh_drv_probe success\n");
	return 0;
}

/*
 * .remove
 */
static int ehci_suh_drv_remove(struct platform_device *device)
{
	struct usb_hcd *hcd;

	hcd = platform_get_drvdata(device);
	usb_remove_hcd(hcd);
	iounmap(hcd->regs);
	release_mem_region(hcd->rsrc_start, hcd->rsrc_len);
	usb_put_hcd(hcd);
	ehci_suh_stop(device);
	platform_set_drvdata(device, NULL);
	dev_dbg(&device->dev,"ehci_suh_drv_remove success\n");
	return 0;
}
#ifdef  CONFIG_PM
/*
 * .suspend
 */
static int ehci_suh_drv_suspend(struct platform_device *device, pm_message_t message)
{
	struct usb_hcd *hcd;
	struct ehci_hcd *ehci;
	unsigned long flags;

	hcd = platform_get_drvdata(device);
	ehci = hcd_to_ehci(hcd);

	dev_dbg(&device->dev,"message.event=%X\n",message.event);
	switch ( message.event ) {
	case PM_EVENT_PRETHAW:
	/* System Depend Process Start */
	ehci_suh_start(device);
	/* System Depend Process End */
		break;
	case PM_EVENT_FREEZE:
	case PM_EVENT_SUSPEND:
	case PM_EVENT_HIBERNATE:
		break;
	default:
		dev_dbg(&device->dev,"message.event fail\n");
		return -EINVAL;
	}

	spin_lock_irqsave(&ehci->lock, flags);

	dev_dbg(&device->dev,"hcd->state=%X\n",hcd->state);
	if ( hcd->state != HC_STATE_SUSPENDED ) {
		spin_unlock_irqrestore(&ehci->lock, flags);
		if ( message.event == PM_EVENT_PRETHAW ) {
			/* System Depend Process Start */
			ehci_suh_stop(device);
			/* System Depend Process End */
		}
		dev_dbg(&device->dev,"hcd->state fail\n");
		return -EINVAL;
	}

	if ( message.event != PM_EVENT_PRETHAW ) {
		ehci_writel(ehci, 0, &ehci->regs->intr_enable);
		ehci_readl(ehci, &ehci->regs->intr_enable);

		clear_bit(HCD_FLAG_HW_ACCESSIBLE, &hcd->flags);

		spin_unlock_irqrestore(&ehci->lock, flags);

		/* System Depend Process Start */
		ehci_suh_stop(device);
		/* System Depend Process End */

		dev_dbg(&device->dev,"ehci_suh_drv_suspend success\n");
	} else {
		set_bit(HCD_FLAG_HW_ACCESSIBLE, &hcd->flags);

		/* System Depend Process Start */
		ehci_suh_set_register(device,(u32)ehci->regs);
		/* System Depend Process End */

		ehci_halt(ehci);
		ehci_reset(ehci);

		clear_bit(HCD_FLAG_HW_ACCESSIBLE, &hcd->flags);

		spin_unlock_irqrestore(&ehci->lock, flags);

		/* System Depend Process Start */
		ehci_suh_stop(device);
		/* System Depend Process End */

		dev_dbg(&device->dev,"ehci_suh_drv_suspend success\n");
	}
	return 0;
}

/*
 * .resume
 */
static int ehci_suh_drv_resume(struct platform_device *device)
{
	struct usb_hcd *hcd;
	struct ehci_hcd *ehci;
	unsigned long flags;
	int mask;

	hcd = platform_get_drvdata(device);
	ehci = hcd_to_ehci(hcd);
	mask = INTR_MASK;

	/* System Depend Process Start */
	ehci_suh_start(device);
	/* System Depend Process End */

	spin_lock_irqsave(&ehci->lock, flags);

	set_bit(HCD_FLAG_HW_ACCESSIBLE, &hcd->flags);

	dev_dbg(&device->dev,"ehci->regs->configured_flag = %X\n",
				ehci_readl(ehci, &ehci->regs->configured_flag));
	if (ehci_readl(ehci, &ehci->regs->configured_flag) == FLAG_CF) {
		if ( hcd->self.root_hub->do_remote_wakeup == 0) {
			mask &= ~STS_PCD;
		}
		ehci_writel(ehci, mask, &ehci->regs->intr_enable);
		ehci_readl(ehci, &ehci->regs->intr_enable);

		spin_unlock_irqrestore(&ehci->lock, flags);

		dev_dbg(&device->dev,"ehci_suh_drv_resume success\n");
	} else {
		ehci_halt(ehci);
		ehci_reset(ehci);

		if ( ehci->reclaim != 0 ) {
			end_unlink_async(ehci);
		}
		ehci_work(ehci);

		ehci_writel(ehci, ehci->command, &ehci->regs->command);
		ehci_writel(ehci, FLAG_CF, &ehci->regs->configured_flag);
		ehci_readl(ehci, &ehci->regs->command);

		ehci_port_power(ehci, 1);

		spin_unlock_irqrestore(&ehci->lock, flags);

		usb_root_hub_lost_power(hcd->self.root_hub);

		dev_dbg(&device->dev,"ehci_suh_drv_resume success\n");
	}
	return 0;
}
#else
#define ehci_suh_drv_suspend NULL
#define ehci_suh_drv_resume  NULL
#endif
