/*
 * ohci-suh.c
 *
 * OHCI HCD (Host Controller Driver) for USB.
 *
 * Copyright 2012 Sony Corporation.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Suite 500, Boston, MA 02110-1335 USA.
 *
 */
/*----------------------------------------------------------------------*
 * header
 *----------------------------------------------------------------------*/
#include <linux/platform_device.h>
#include <linux/io.h>

#include "ohci-suh.h"
#ifdef CONFIG_POSTEDWRITE_WORKAROUND
#include "pos-suh_def.h"
#include <mach/postedwrite_sync.h>
#endif

/*----------------------------------------------------------------------*
 * extern
 *----------------------------------------------------------------------*/
extern int usb_disabled(void);

/*----------------------------------------------------------------------*
 * function
 *----------------------------------------------------------------------*/
static void ohci_suh_start_clock(void);
static void ohci_suh_stop_clock(void);
static void ohci_suh_start_hc(struct platform_device *);
static void ohci_suh_stop_hc(struct platform_device *);
static void ohci_suh_init(struct platform_device *);
static int __devinit ohci_suh_start (struct usb_hcd *);
static int usb_hcd_suh_probe(const struct hc_driver *,
			struct platform_device *);
static void usb_hcd_suh_remove(struct usb_hcd *,
				struct platform_device *);
static int ohci_suh_drv_probe(struct platform_device *);
static int ohci_suh_drv_remove(struct platform_device *);
#ifdef CONFIG_PM
static int ohci_suh_drv_resume(struct platform_device *);
static int ohci_suh_drv_suspend(struct platform_device *, pm_message_t);
#endif
static void ohci_suh_mon(struct platform_device *, __u32 __iomem *);

/*----------------------------------------------------------------------
 * user implementation part
 *----------------------------------------------------------------------*/

/*
 * ohci_suh_start_clock - called when the device glued to the bus or in probe.
 * required for operation -- clock-on and reset-off
 * but no used now.
 */
static void ohci_suh_start_clock(void)
{
	/* clock on */
}

/*
 * ohci_suh_stop_clock - called when the device removed from the bus
 * or in probe error.
 * required for operation -- clock-off and reset-on
 * but no used now.
 */
static void ohci_suh_stop_clock(void)
{
	/* clock off */
}
/*----------------------------------------------------------------------
 * implementation part for HCD
 *----------------------------------------------------------------------*/

/*
 * ohci_suh_start_hc - makes preparations for operation of OHCI
 *  host controller.
 * it keeps the reset-on of OHCI controller.
 */
static void ohci_suh_start_hc(struct platform_device *pdev)
{
	struct usb_hcd *hcd = platform_get_drvdata(pdev);
	struct ohci_regs __iomem *regs = hcd->regs;

	dev_dbg(&pdev->dev, "start\n");

	/*
	 * Start the USB clocks.
	 */
	ohci_suh_start_clock();

	/*
	 * The USB host controller must remain in reset.
	 */
	writel(0, &regs->control);
}

/*
 * ohci_suh_stop_hc - stops operations of OHCI host controller
 * after it keeps the reset-on of OHCI controller.
 */
static void ohci_suh_stop_hc(struct platform_device *pdev)
{
	struct usb_hcd *hcd = platform_get_drvdata(pdev);
	struct ohci_regs __iomem *regs = hcd->regs;

	dev_dbg(&pdev->dev, "stop\n");

	/*
	 * Put the USB host controller into reset.
	 */
	writel(0, &regs->control);

	/*
	 * Stop the USB clocks.
	 */
	ohci_suh_stop_clock();
}
/*
 * ohci_suh_init - setting and config of HC specific registers.
 */
static void ohci_suh_init(struct platform_device *pdev)
{
	struct usb_hcd *hcd = platform_get_drvdata(pdev);

	/* nothing to do for setting now */

	/* monitor */
	ohci_suh_mon(pdev, (__hc32 __iomem *)(hcd->regs));

}

/*
 * ohci_suh_mon - dump of HC specific registers.
 */
static void ohci_suh_mon(struct platform_device *pdev, __hc32 __iomem *base)
{

	/* REG06 */
	dev_dbg(&pdev->dev, " IP-vendorSpec INSNREG06: 0x%x \n",
	 __raw_readl(base + SUH_OFFSET_INSNREG06));

	/* REG07 */
	dev_dbg(&pdev->dev, " IP-vendorSpec INSNREG07: 0x%x \n",
	 __raw_readl(base + SUH_OFFSET_INSNREG07));

}
/*
 * ohci_suh_start - initialize and excute ohci.
 * Since ".start" of "platform_bus" and and "ohci_run()" differ
 * in an argument type, they cannot call "ohci_run" directly
 * like "ehci_run" in EHCI.
 */
static int __devinit ohci_suh_start(struct usb_hcd *hcd)
{
	struct	ohci_hcd *ohci = hcd_to_ohci (hcd);
	int	ret;

	if ((ret = ohci_init(ohci)) < 0)
		return ret;

	if ((ret = ohci_run(ohci)) < 0) {
		/* ohci_run failed */
		ohci_err(ohci, "can't start %s", hcd->self.bus_name);
		ohci_stop(hcd);	/* free memory got in ohci_int */
		return ret;
	}
	return 0;
}

/*----------------------------------------------------------------------*
 * OPS structure
 *----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*
 * hc_driver:
 *----------------------------------------------------------------------*/
static const struct hc_driver ohci_suh_hc_driver = {
	.description =		hcd_name,
	.product_desc =		"SUH OHCI",
	.hcd_priv_size =	sizeof(struct ohci_hcd),

	/*
	 * generic hardware linkage
	 */
	.irq =			ohci_irq,
	.flags =		HCD_USB11 | HCD_MEMORY,

	/*
	 * basic lifecycle operations
	 */
	.start =		ohci_suh_start,	/* call local function */
	.stop =			ohci_stop,
	.shutdown =		ohci_shutdown,

	/*
	 * managing i/o requests and associated device resources
	 */
	.urb_enqueue =		ohci_urb_enqueue,
	.urb_dequeue =		ohci_urb_dequeue,
	.endpoint_disable =	ohci_endpoint_disable,

	/*
	 * scheduling support
	 */
	.get_frame_number =	ohci_get_frame,

	/*
	 * root hub support
	 */
	.hub_status_data =	ohci_hub_status_data,
	.hub_control =		ohci_hub_control,
#ifdef CONFIG_PM
	.bus_suspend =		ohci_bus_suspend,
	.bus_resume =		ohci_bus_resume,
#endif
	.start_port_reset =	ohci_start_port_reset,
};
/*
 * usb_hcd_suh_probe - initialize SUH-based HCDs
 * Context: !in_interrupt()
 *
 * this is the main function of probe called from platform_bus.
 * this function gets resources of the platform_device
 * (e.g. memory or irq) and makes HCD structure, and then
 * sets clock on for USB, allocates of HC specific register area,
 * etc. HC is started after preparation is completed.
 */
static int usb_hcd_suh_probe(const struct hc_driver *driver,
			struct platform_device *pdev)
{
	int	retval;
	struct	resource *rsrc;
	int 	irq;
	struct usb_hcd *hcd = NULL;

	if(usb_disabled()){
		return -ENODEV;
	}
	if (pdev->num_resources != 2) { /* resource type is either MEM or IRQ */
		dev_err(&pdev->dev,"hcd probe: invalid num_resources: %d\n",
			pdev->num_resources);
		return -ENODEV;
	}

	rsrc = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if(!rsrc){
		dev_err(&pdev->dev,"hcd probe: invalid num_resources no MEM: %d\n",
			pdev->num_resources);
		return -ENODEV;
	}
	irq = platform_get_irq(pdev, 0);
	if(irq < 0){
		dev_err(&pdev->dev,"hcd probe: invalid num_resources no IRQ: %d\n",
			pdev->num_resources);
		return -ENODEV;
	}

	/* create structure of usb_hcd */
	hcd = usb_create_hcd (&ohci_suh_hc_driver, &pdev->dev, dev_name(&pdev->dev));
	if (!hcd){
		return -ENOMEM;
	}
	/* start point of register area */
	hcd->rsrc_start = rsrc->start;

	/* size of register area */
	hcd->rsrc_len = rsrc->end - rsrc->start + 1;

	/* request assign of memory region */
	if (!request_mem_region(hcd->rsrc_start, hcd->rsrc_len, pdev->name)) {
		dev_err(&pdev->dev,"request_mem_region failed\n");
		retval = -EBUSY;
		goto err1;
	}
	/* assign of virtual memory area to access physical memory area */
	hcd->regs = ioremap_nocache(hcd->rsrc_start, hcd->rsrc_len);
	if (!hcd->regs) {
		dev_err(&pdev->dev,"ioremap failed\n");
		retval = -ENOMEM;
		goto err2;
	}

	/* handle clock start */
	ohci_suh_start_hc(pdev);

	/* set hw-specific function */
	ohci_suh_init(pdev);

	/* drvdata already has been set in usb_create_hcd but here set again */
	platform_set_drvdata(pdev, hcd);

	ohci_hcd_init(hcd_to_ohci(hcd));

	retval = usb_add_hcd(hcd, irq, IRQF_DISABLED | IRQF_SHARED);
	if (retval == 0){
		/* normal end */
		return retval;
	}
	/* Error handling */
	ohci_suh_stop_hc(pdev);

	iounmap(hcd->regs);

 err2:
	release_mem_region(hcd->rsrc_start, hcd->rsrc_len);

 err1:
	usb_put_hcd(hcd);

	return retval;
}

/*
 * usb_hcd_suh_remove - shutdown processing for SUH-based HCDs
 * @dev: USB Host Controller being removed
 * Context: !in_interrupt()
 *
 * memories for a HCD structure is released , stop a clock for USB
 * and HC register area is unmapped.
 */
static void usb_hcd_suh_remove(struct usb_hcd *hcd,
				struct platform_device *pdev)
{
	usb_remove_hcd(hcd);
	/* hc stop */
	ohci_suh_stop_hc(pdev);
	iounmap(hcd->regs);
	release_mem_region(hcd->rsrc_start, hcd->rsrc_len);
	usb_put_hcd(hcd);
	/* reserved : for clock release */
	platform_set_drvdata(pdev, NULL);
}

/*
 * ohci_suh_drv_probe - linked to a probe function of "platform_bus".
 */
static int ohci_suh_drv_probe(struct platform_device *pdev)
{

	device_init_wakeup(&pdev->dev, 1);
	return usb_hcd_suh_probe(&ohci_suh_hc_driver, pdev);
}

/*
 * ohci_suh_drv_remove - linked to a remove function of "platform_bus".
 */
static int ohci_suh_drv_remove(struct platform_device *pdev)
{
	device_init_wakeup(&pdev->dev, 0);
	usb_hcd_suh_remove(platform_get_drvdata(pdev), pdev);

	return 0;
}

#ifdef CONFIG_PM
/* ohci_suh_drv_suspend - linked to a suspend function of "platform_bus".
 * processing is classified according to a message given.
 * it is only enable in case CONFIG_PM exists.
 * -- policy --
 * 1)suspend(sus to ram):
 *  keep clock on and expect power saveed by PHY & LINK
 *  suspend action.
 *  (if no device is attached, the port is suspended.)
 * 2)hibernation(sus to disk):
 *  when entering : device is powered off
 *  when exitting  : device is powered on
 */
static int ohci_suh_drv_suspend(struct platform_device *pdev,
				pm_message_t mesg)
{
	dev_dbg(&pdev->dev, "SUH: platformbus suspend in state:[%d] \n", mesg.event);

	switch (mesg.event) {
		case PM_EVENT_FREEZE:
			ohci_suh_stop_hc(pdev);
			/* reserved to do for clock off */
			break;
		case PM_EVENT_SUSPEND:
		case PM_EVENT_HIBERNATE:
			/* nothing to handle clock */
			break;
		case PM_EVENT_PRETHAW:
			/* reserved to do for clock on */

			/* handle clock start */
			ohci_suh_start_hc(pdev);
			break;
		default:
			dev_err(&pdev->dev,"unsupported mesg!! \n");
	}
	return 0;
}

/*
 * ohci_suh_drv_resume - linked to a resume function of "platform_bus".
 * processing is classified according to a message given.
 * it is only enable in case CONFIG_PM exists.
 * -- policy --
 * 1)suspend(suspend to ram):
 *  nothing to do,
 *  but needed to restore clock if clock was handled in suspend.
 * 2)hibernation(suspend to disk):
 *  nothing to do
 */
static int ohci_suh_drv_resume(struct platform_device *pdev)
{
	dev_dbg(&pdev->dev, "SUH: platformbus resume \n");

	return 0;
}
#else
#define ohci_suh_drv_suspend NULL
#define ohci_suh_drv_resume  NULL
#endif

MODULE_ALIAS("platform:ohci-suh");

/*
 * platform_driver -- the structure for OHCI as a platform_bus driver
 * driver.name: identifier used to refer to platform_device
 */
static struct platform_driver ohci_suh_platform_driver = {
	.probe		= ohci_suh_drv_probe,
	.remove		= ohci_suh_drv_remove,
	.shutdown	= usb_hcd_platform_shutdown,
	.suspend	= ohci_suh_drv_suspend,
	.resume		= ohci_suh_drv_resume,
	.driver		= {
		.name	= "ohci-suh",
		.owner	= THIS_MODULE,
	},
};
